/**
 * An implementation of various Load Balancing schemes, to examine the concepts of Isoefficiency.
 *
 * MPI Tags:
 *  * 1 - Termination message from main.
 *  * 2 - Work request message.
 *  * 3 - Status update message.
 */


// System Import Headers.
#include <ctype.h>
#include <limits.h>
#include <mpi.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/sysinfo.h>
#include <unistd.h>

// User Import Headers.
#ifndef user_headers
    #define user_headers
    #include "load_balance_schemes.h"
    #include "structs.h"
    #include "terminal_commands.h"
#endif


// Method Declaration.
void validate_args();       // Validates passed user args.
void display_max_types();   // Displays a few data type max values.
void run_program();         // Main program logic. Handles various Load Balance schemas.


// Global Variables.
int total_processors;
int process_rank;
int seconds_per_load;
int total_loads;


/**
 * Program's main. Initialized and runs program.
 */
int main(int argc, char* argv[]) {

    // Initialize thread communication.
    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &total_processors);
    MPI_Comm_rank(MPI_COMM_WORLD, &process_rank);

    if (process_rank == 0) {
        printf("Initializing program.\n");
        printf("\n");
    }

    // Validate program args.
    validate_args(argc, argv);

    // If we got this far, then args are valid. Proceed with program.
    // total_processors = get_nprocs();

    // Display args.
    if (process_rank == 0) {
        printf("System & Program Variables:\n");
        printf("    %i Total Loads\n", total_loads);
        printf("    %i Seconds Per Load\n", seconds_per_load);
        printf("\n");
        printf("    Total Expected work = TotalLoads * SecondsPerLoad\n");
        printf("                        = %i * %i\n", total_loads, seconds_per_load);
        int seconds_active_work = total_loads * seconds_per_load;
        printf("                        = %i Seconds Active Work\n", seconds_active_work);
        printf("                        = %i Minutes Active Work\n", (seconds_active_work / 60));
        printf("\n");
        printf("    %i Total System Processors\n", get_nprocs());
        printf("    %i Processors in Use\n", total_processors);
        printf("\n");
        printf("    Expected total execution time with ideal processor usage (ignoring unused processors):\n");
        printf("    (Work / Processor Count) = %i Minutes\n", (seconds_active_work / 60 / total_processors));
        printf("\n");
    }

    // Wait for all processes to synchronize.
    MPI_Barrier(MPI_COMM_WORLD);

    // Run program.
    // display_max_types();
    run_program();

    if (process_rank == 0) {
        printf("\n");
        printf("Terminating program.\n");
    }

    // Close thread communication.
    MPI_Finalize();

    exit(0);
}


/**
 * Validate program args.
 */
void validate_args(int argc, char* argv[]) {

    if (argc == 3) {
        // Expected number of params. Validate passed args.

        // Validate "seconds_per_load" value. Should be between 1 and 10.
        seconds_per_load = strtol(argv[1], NULL, 10);
        if ((seconds_per_load < 1) || (seconds_per_load > 10)) {
            if (process_rank == 0) {
                printf("Arg1 (seconds_per_load) should be int between 1 and 10.\n");
                printf("\n");
                printf("Terminating program.\n");
            }
            exit(1);
        }

        // Validate "total_loads" value. Should be between 100 and 10,000.
        total_loads = strtol(argv[2], NULL, 10);
        if ((total_loads < 10) || (total_loads > 10000)) {
            if (process_rank == 0) {
                printf("Arg2 (total_loads) should be int between 10 and 10,000.\n");
                printf("\n");
                printf("Terminating program.\n");
            }
            exit(1);
        }

    } else if (argc > 3) {
        // Too many args. Error.
        if (process_rank == 0) {
            printf("Too many args passed. Got %i. Expected 2.\n", (argc - 1));
            printf("\n");
            printf("Terminating program.\n");
        }
        exit(1);

    } else {
        // Too few args. Error.
        if (process_rank == 0) {
            printf("Too few args passed. Got %i. Expected 2.\n", (argc - 1));
            printf("\n");
            printf("Terminating program.\n");
        }
        exit(1);
    }
}


/**
 * Helper function to display some max values for system.
 */
void display_max_types() {
    printf("\n\n");

    printf("Int max is %i.\n", INT_MAX);
    printf("Unsigned Int max is %u.\n", UINT_MAX);
    printf("Long Max is %ld.\n", LONG_MAX);
    printf("Unsigned Long max is %lu.\n", ULONG_MAX);

    printf("\n\n");
}


/**
 * Main program logic.
 * Entry point for running various Load Balancing schemes.
 */
void run_program() {
    if (process_rank == 0) {
        printf("Running core program logic.\n");
        printf("\n");
    }

    // Wait for all processes to synchronize.
    MPI_Barrier(MPI_COMM_WORLD);

    thread_struct* thread_args_ptr = initialize_thread_struct(
        total_processors, seconds_per_load, total_loads, process_rank
    );

    arr_run(thread_args_ptr);

    free_thread_struct(thread_args_ptr);

    // Wait for all processes to synchronize.
    // MPI_Barrier(MPI_COMM_WORLD);

    if (process_rank == 0) {
        printf("\n");
        printf("Core program logic complete.\n");
    }
}
